;;; swing.lisp
;;;
;;; Copyright (C) 2008-2009 Alessio Stalla
;;;
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
;;;
;;; As a special exception, the copyright holders of this library give you
;;; permission to link this library with independent modules to produce an
;;; executable, regardless of the license terms of these independent
;;; modules, and to copy and distribute the resulting executable under
;;; terms of your choice, provided that you also meet, for each linked
;;; independent module, the terms and conditions of the license of that
;;; module.  An independent module is a module which is not derived from
;;; or based on this library.  If you modify this library, you may extend
;;; this exception to your version of the library, but you are not
;;; obligated to do so.  If you do not wish to do so, delete this
;;; exception statement from your version.

(defpackage :snow-swing
  (:use :common-lisp :snow :java :ext :named-readtables #+snow-cells :cells)
  (:shadowing-import-from :snow #:make-dialog-prompt-stream #:*gui-backend*))

(in-package :snow-swing)

(defmacro defimpl (name args &body body)
  `(defimplementation ,name (*gui-backend* :swing) ,args
     ,@body))

(defun make-action-listener (obj)
  (if (or (functionp obj) (symbolp obj))
      (jmake-proxy "java.awt.event.ActionListener"
		   (let ((env (or snow::*dynamic-environment*
				  (snow::capture-dynamic-environment))))
		     (lambda (this method-name event)
		       (declare (ignore this method-name))
		       (snow::with-dynamic-environment (env)
			 (let ((snow::*dynamic-environment* env))
			   (funcall obj event))))))
      obj)) ;This allows to use a native Java action listener

(defimpl make-layout-manager (widget layout &rest args)
  (if (typep layout 'java-object)
      layout
      (ecase layout
	((or :default :mig) (apply #'new "net.miginfocom.swing.MigLayout" args))
	(:box (new "javax.swing.BoxLayout"
		   (if (jinstance-of-p widget "javax.swing.JFrame")
		       (invoke "getContentPane" widget)
		       widget)
		   (ecase (car args)
		     (:x (jfield "javax.swing.BoxLayout" "X_AXIS"))
		     (:y (jfield "javax.swing.BoxLayout" "Y_AXIS")))))
	(:flow (new "java.awt.FlowLayout"))
	(:border (new "java.awt.BorderLayout"))
	((nil) nil))))

(defimpl (setf layout-manager) (lm widget)
  (setf (widget-property widget :layout) lm))

(defconstant +add-to-container+ (jmethod "java.awt.Container" "add" "java.awt.Component"))

(defconstant +add-to-container-with-constraints+ (jmethod "java.awt.Container" "add" "java.lang.String" "java.awt.Component"))

(defimpl call-in-gui-thread (fn)
  (jstatic "invokeLater" "javax.swing.SwingUtilities"
	   (new "snow.FunctionRunnable" fn)))

;;Base API implementation
(defimpl add-child (child &optional (parent *parent*) layout-constraints)
  (if layout-constraints
      (jcall +add-to-container-with-constraints+
	     parent
	     layout-constraints
	     child)
      (jcall +add-to-container+ parent child)))

(defimpl (setf widget-enabled-p) (value widget)
  (setf (widget-property widget :enabled) value))

(defimpl widget-enabled-p (widget)
  (widget-property widget :enabled))

(defimpl (setf widget-visible-p) (value widget)
  (setf (widget-property widget :visible) value))

(defimpl widget-visible-p (widget)
  (widget-property widget :visible))

(defimplementation (setf widget-location) (*gui-backend* :swing) (value widget)
  (invoke "setLocation" widget (aref value 0) (aref value 1)))

(defimpl (setf widget-size) (value widget)
  (invoke "setSize" widget (realpart value) (imagpart value)))

(defimpl dispose (obj)
  (invoke "dispose" obj))

(defimpl show (obj)
  (invoke "show" obj))

(defimpl hide (obj)
  (invoke "hide" obj))

;;; --- Widgets --- ;;;

;Frames and dialogs
(defimpl snow::make-frame (&key menu-bar title visible-p on-close
			   &allow-other-keys)
  (let ((f (new "javax.swing.JFrame")))
    (set-widget-properties f
      :title title 
      :visible (jbool visible-p))
    (when menu-bar
      (setf (widget-property f "JMenuBar") menu-bar))
    (when on-close
      (let ((on-close
	     (case on-close
	       ((#'ext:exit 'ext:exit :exit)
		(lambda (evt)
		  (declare (ignore evt))
		  (ext:exit)))
	       (t on-close))))
	(invoke "addWindowListener" f (new "snow.swing.WindowListener"
					   nil nil on-close nil nil nil nil))))
    f))

(defimpl snow::make-dialog (&key parent title modal-p visible-p &allow-other-keys)
  (let ((d (new "javax.swing.JDialog"
		parent
		(if modal-p
		    (jfield "java.awt.Dialog$ModalityType" "APPLICATION_MODAL")
		    (jfield "java.awt.Dialog$ModalityType" "MODELESS")))))
    (set-widget-properties d
      :title title)
    d))

(defimpl pack (window)
  (jcall (jmethod "java.awt.Window" "pack") window)
  window)

(defun setup-button (btn text on-action)
  (when text
    (setf (widget-property btn :text) text))
  (when on-action
    (invoke "addActionListener"
	    btn
	    (make-action-listener on-action))))

;;Menus
(defimpl snow::make-menu-bar (&key &allow-other-keys)
  (new "javax.swing.JMenuBar"))

(defimpl snow::make-menu (&key text &allow-other-keys)
  (if text
      (new "javax.swing.JMenu" text)
      (new "javax.swing.JMenu")))

(defimpl snow::make-menu-item (&key text on-action &allow-other-keys)
  (let ((m (new "javax.swing.JMenuItem")))
    (setup-button m text on-action)
    m))

;;Panels
(defimpl snow::make-panel (&key &allow-other-keys)
  (new "javax.swing.JPanel"))

(defimpl snow::make-tabs (&key (wrap t) (tab-placement :top) &allow-other-keys)
  (let ((tabs (new "javax.swing.JTabbedPane")))
    (invoke "setTabLayoutPolicy" tabs
	    (if wrap
		#.(jfield "javax.swing.JTabbedPane" "WRAP_TAB_LAYOUT")
		#.(jfield "javax.swing.JTabbedPane" "SCROLL_TAB_LAYOUT")))
    (invoke "setTabPlacement" tabs
	    (ecase tab-placement
	      (:top #.(jfield "javax.swing.JTabbedPane" "TOP"))
	      (:bottom #.(jfield "javax.swing.JTabbedPane" "BOTTOM"))
	      (:left #.(jfield "javax.swing.JTabbedPane" "LEFT"))
	      (:right #.(jfield "javax.swing.JTabbedPane" "RIGHT"))))
    tabs))

(defimpl snow::make-scroll-panel (view)
  (let ((p (new "javax.swing.JScrollPane")))
    (setf (scroll-panel-view p) view)
    p))

(defimpl snow::scroll-panel-view (self)
  (jproperty-value self "viewportView"))

(defimpl (setf snow::scroll-panel-view) (view self)
  (setf (jproperty-value self "viewportView") view))

(defimpl snow::make-split-panel (child1 child2
				 &key (orientation :horizontal) smoothp)
  (new "javax.swing.JSplitPane"
       (ecase orientation
	 ((or :horizontal :h nil)
	  #.(jfield "javax.swing.JSplitPane" "HORIZONTAL_SPLIT"))
	 ((or :vertical :v)
	  #.(jfield "javax.swing.JSplitPane" "VERTICAL_SPLIT")))
       (jbool smoothp)
       child1
       child2))

;Buttons
(defimpl snow::make-button (&key text on-action &allow-other-keys)
  (let ((btn (new "javax.swing.JButton")))
    (setup-button btn text on-action)
    btn))

(defimpl snow::make-check-box (&key text selected-p &allow-other-keys)
  (let ((btn (new "javax.swing.JCheckBox")))
    (when text
      (setf (widget-property btn :text) text))
    (setf (widget-property btn :selected)
	  (if selected-p selected-p (jbool nil)))
    btn))

;Misc
(defconstant +swingconstant-vertical+ 1) ; it should be something like (jmethod "javax.swing.SwingConstants" "VERTICAL")
(defimpl snow::make-progress-bar (&key value orientation (paint-border t) progress-string &allow-other-keys)
  (let ((pbar (new "javax.swing.JProgressBar")))
    (when value
      (setf (widget-property pbar :value) value))
   (when orientation
     (setf (widget-property pbar :orientation) +swingconstant-vertical+))
   (when (not paint-border)
     (setf (widget-property pbar :border-painted) (jbool nil)))
   (when progress-string
     (setf (widget-property pbar :string-painted) (jbool t))
     (setf (widget-property pbar :string) progress-string))
   pbar))

;Text
(defimpl snow::make-label (&key text &allow-other-keys)
  (let ((lbl (new "javax.swing.JLabel")))
    (when text
      (setf (widget-property lbl :text) text))
    lbl))

(defimpl snow::make-text-field (&key text &allow-other-keys)
  (let ((field (new "javax.swing.JTextField")))
    (when text
      (setf (widget-property field :text) text))
    field))

(defimpl snow::make-text-area (&key text &allow-other-keys)
  (let ((text-area (new "javax.swing.JTextArea")))
    (when text
      (setf (widget-property text-area :text) text))
    text-area))

(defun snow::make-dialog-prompt-stream () ;;todo!!
  (new "snow.SwingDialogPromptStream"))

;;Lists
(defun make-list-model (list)
  (new "snow.list.ConsListModel" list))

(defimpl snow::make-list-widget (&key model prototype-cell-value selected-index
				      (cell-renderer (new "snow.list.ConsListCellRenderer"))
				 &allow-other-keys)
  (let ((list (new "javax.swing.JList")))
    (when model (setf (widget-property list :model) model))
    (setf (widget-property list :cell-renderer) cell-renderer)
    (setf (widget-property list :prototype-cell-value) prototype-cell-value)
    (when selected-index
      (setf (widget-property list :selected-index) selected-index))
    list))

;;Trees
(defun make-tree-model (list)
  (new "snow.tree.ConsTreeModel" list))

(defimpl snow::make-tree (&key model (cell-renderer (new "snow.tree.ConsTreeCellRenderer"))
				 &allow-other-keys)
  (let ((tree (new "javax.swing.JTree")))
    (when model (setf (widget-property tree :model) model))
    (setf (widget-property tree :cell-renderer) cell-renderer)
    tree))

;;REPL
(defimpl snow::make-gui-repl (&key dispose-on-close environment)
  (let ((text-area (new "javax.swing.JTextArea"))
	(repl-doc (new "snow.swing.ConsoleDocument"
		       (compile nil
				`(lambda ()
				   (let (,@environment)
				     ;;safe: *debugger-hook* is rebound
				     (install-graphical-debugger)
				     (top-level::top-level-loop)))))))
    (setf (widget-property text-area :document) repl-doc)
    (invoke "setupTextComponent" repl-doc text-area)
    (when dispose-on-close
      (invoke "disposeOnClose" repl-doc dispose-on-close))
    text-area))

(defimpl snow::dispose-gui-repl (repl)
  (invoke "dispose" (widget-property repl :document)))
